@@ -17,41 +17,39 @@ Implementation details:
 Optional:
 
 * Validating PRODUCT_CLASSES_ALLOWED
-* Validating upstream, product and params in code
 * Using a client parameter
 
-NOTE: Params trigger different data output (and should make tasks outdated),
-Tasks constructor args (such as chunksize in SQLDump) should not change
-the output, hence shoulf not make tasks outdated
+Task execution model:
 
-Task Status lifecycle:
+Upon instantiation, all tasks start with a WaitingRender status. On successful
+render, they move to Skipped if they are up-to-date and all upstream
+dependencies are up-to-date as well, WaitingExecution if they
+are outdated but upstream dependencies are up-to-date or WaitingUpstream, if
+any upstream dependency is outdated.
 
-* They all start on WaitingRender (set in __init__)
-* (dag.render() is called)
-* Tasks move to:
-    * If sucessful render: TaskStatus.Skipped, WaitingExecution or
-        WaitingUpstream
-    * If failed render: ErroredRender or AbortedRender
-    * This happens in Task.render()
-    * AbortedRender is not set directly, but by propagating downstream
-* (dag.build() is called)
-* Check that all TaskStatus are either WaitingExecution, WaitingUpstream
-    or Skipped (currently done in Executor)
-* Once dag.build() starts, the executor must call task.build() inside a
-    try-catch statement to set status to Executed or Errored
-* When tasks are set to Errored, Abort is set to downstream tasks
+If render fails, they are marked as ErroredRender and downstream dependencies
+are marked as AbortedRender.
 
-TODO: describe BrokenProcesssPool
+When a Task is built, and succeeds, the on_finish hook is executed, if
+on_finish succeeds as well, the Task is marked as Executed. If either
+the task fails or on_finish fails, the Task is marked as Errored.
 
+When a task is marked as Executed, is saves metadata in its Product,
+if the product registered a prepare_metadata hook, the metadata to save
+can be modified before is saved.
+
+When a Task is marked as Errored, on_failure hook is executed and downstream
+dependencies are marked as Aborted.
+
+
+TODO: describe BrokenProcesssPool status
 """
-import traceback
 import abc
 import logging
-import warnings
 from datetime import datetime
 from ploomber.products import Product, MetaProduct
 from ploomber.dag.DAG import DAG
-from ploomber.exceptions import TaskBuildError
+from ploomber.exceptions import TaskBuildError, DAGBuildEarlyStop
 from ploomber.tasks.TaskGroup import TaskGroup
 from ploomber.constants import TaskStatus
 from ploomber.tasks.Upstream import Upstream
@@ -119,27 +117,26 @@ class Task(abc.ABC):
         pass
 
     @abc.abstractmethod
-    def _init_source(self, source):
+    def _init_source(self, source, kwargs):
         pass
 
     def __init__(self, source, product, dag, name=None, params=None):
         self._params = Params(params)
-        self._source = self._init_source(source)
+
+        kwargs = dict(hot_reload=dag._params.hot_reload)
+        self._source = self._init_source(source, kwargs)
 
         if name is None:
-            # works with pathlib.Path and ploomber.Placeholder
-            if hasattr(source, 'name'):
-                self._name = source.name
-            # works with python functions
-            elif hasattr(source, '__name__'):
-                self._name = source.__name__
-            else:
-                raise AttributeError('Task name can only be None if the '
-                                     'source object has a "name" attribute: '
-                                     'Placeholder objects (returned from '
-                                     'SourceLoader) or pathlib.Path objects. '
-                                     'Python functions also work since they '
-                                     'have a "__name__" attribute')
+            # use name inferred from the source object
+            self._name = self._source.name
+
+            if self._name is None:
+                raise AttributeError('Task name can only be None if a name'
+                                     'can be inferred from the source object. '
+                                     'This only works when the task receives a'
+                                     'pathlib.Path, when using SourceLoader '
+                                     'or in PythonCallable. Pass a value '
+                                     'explicitely.')
         else:
             self._name = name
 
@@ -280,16 +277,41 @@ class Task(abc.ABC):
         self._on_finish = value
 
     def _run_on_finish(self):
+        # run on finish first, if this fails, we don't want to save metadata
         if self.on_finish:
             kwargs = callback_check(self.on_finish,
                                     self._available_callback_kwargs)
             try:
                 self.on_finish(**kwargs)
             except Exception as e:
+                msg = ('Exception when running on_finish '
+                       'for task "{}": {}'
+                       .format(self.name, e))
+                self._logger.exception(msg)
                 self.exec_status = TaskStatus.Errored
-                raise type(e)('Exception when running on_finish '
-                              'for task "{}": {}'
-                              .format(self.name, e)) from e
+                raise type(e)(msg) from e
+
+        self.product._save_metadata(str(self.source))
+
+        # For most Products, it's ok to do this check before
+        # saving metadata, but not for GenericProduct, since the way
+        # exists() works is by checking metadata, so we have to do it
+        # here, after saving metadata
+        if not self.product.exists():
+            if isinstance(self.product, MetaProduct):
+                raise TaskBuildError(
+                    'Error building task "{}": '
+                    'the task ran successfully but product '
+                    '"{}" does not exist yet '
+                    '(task.product.exists() returned False). '
+                    .format(self, self.product))
+            else:
+                raise TaskBuildError(
+                    'Error building task "{}": '
+                    'the task ran successfully but at least one of the '
+                    'products in "{}" does not exist yet '
+                    '(task.product.exists() returned False). '
+                    .format(self, self.product))
 
     @property
     def on_failure(self):
@@ -308,12 +330,7 @@ class Task(abc.ABC):
         if self.on_failure:
             kwargs = callback_check(self.on_failure,
                                     self._available_callback_kwargs)
-            try:
-                self.on_failure(**kwargs)
-            except Exception:
-                tr = traceback.format_exc()
-                warnings.warn('Exception when running on_failure '
-                              'for task "{}". {}'.format(self.name, tr))
+            self.on_failure(**kwargs)
 
     @property
     def on_render(self):
@@ -333,10 +350,12 @@ class Task(abc.ABC):
             try:
                 self.on_render(**kwargs)
             except Exception as e:
+                msg = ('Exception when running on_render '
+                       'for task "{}": {}'
+                       .format(self.name, e))
+                self._logger.exception(msg)
                 self.exec_status = TaskStatus.ErroredRender
-                raise type(e)('Exception when running on_render '
-                              'for task "{}": {}'
-                              .format(self.name, e)) from e
+                raise type(e)(msg) from e
 
     @property
     def exec_status(self):
@@ -354,62 +373,143 @@ class Task(abc.ABC):
         self._logger.debug('Setting "%s" status to %s', self.name, value)
         self._exec_status = value
 
+        # some executors run this in a subprocess, the main process will not
+        # get the new metadata since it was not saved from there, reload
+        if value == TaskStatus.Executed:
+            self.product.metadata.load()
+
         # process might crash, propagate now or changes might not be
         # reflected (e.g. if a Task is marked as Aborted, all downtream
         # tasks should be marked as aborted as well)
-        self._update_downstream_status()
-
         # FIXME: this is inefficient, it is better to traverse
         # the dag in topological order but exclude nodes not affected by
         # this change
+        self._update_downstream_status()
 
-        # TODO: move this to build method? do we really need to run this in
-        # the main process?
-        if value == TaskStatus.Executed:
-            # run on finish first, if this fails, we don't want to save
-            # metadata
-            # Exceptions are *not* silenced here
-            self._run_on_finish()
+    def build(self, force=False, catch_exceptions=True):
+        """Build a single task
 
-            self.product._save_metadata(str(self.source))
+        Although Tasks are primarily designed to execute via DAG.build(), it
+        is possible to do so in isolation. However, this only works if the
+        task does not have any unrendered upstream dependencies, if that's the
+        case, you should call DAG.render() before calling Task.build()
 
-            # NOTE: for most Products, it's ok to do this check before
-            # saving metadata, but not for GenericProduct, since the way
-            # exists() works is by checking metadata, so we have to do it
-            # after saving metadata for it to work
-            if not self.product.exists():
-                raise TaskBuildError('Error building task "{}": '
-                                     'the task ran successfully but product '
-                                     '"{}" does not exist yet '
-                                     '(task.product.exists() returned False)'
-                                     .format(self, self.product))
-
-        elif value == TaskStatus.Errored:
-            # Exceptions here are silenced
-            self._run_on_failure()
-
-    def build(self, force=False, within_dag=False):
-        """build the task
+        Examples
+        --------
+        >>> from pathlib import Path
+        >>> from ploomber import DAG
+        >>> from ploomber.tasks import PythonCallable
+        >>> from ploomber.products import File
+        >>> def fn(product):
+        ...     Path(str(product)).touch()
+        >>> PythonCallable(fn, File('file.txt'), dag=DAG()).build()
 
         Returns
         -------
         dict
             A dictionary with keys 'run' and 'elapsed'
+
+        Raises
+        ------
+        TaskBuildError
+            If the error failed to build because it has upstream dependencies,
+            the build itself failed or build succeded but on_finish hook failed
+
+        DAGBuildEarlyStop
+            If any task or on_finish hook raises a DAGBuildEarlyStop error
         """
-        if within_dag:
-            return self._build(force)
+        if any(t.exec_status == TaskStatus.WaitingRender
+               for t in self.upstream.values()):
+            raise TaskBuildError('Cannot directly build task "{}" as it '
+                                 'has upstream dependencies, call '
+                                 'dag.render() first'.format(self.name))
+
+        # This is the public API for uses who'd to run tasks in isolation,
+        # we have to make sure we clear product cache status, otherwise
+        # this will interfer with other render calls
+        self.render(force=force)
+
+        res = self._build(catch_exceptions=catch_exceptions)
+        self.product._clear_cached_status()
+        return res
+
+    def _build(self, catch_exceptions):
+        """
+        Private API for building DAGs. This is what executors should call.
+        Unlike the public method, this one does not call render, as it
+        should happen via a dag.render() call
+        """
+
+        if not catch_exceptions:
+            res = self._run()
+            self._run_on_finish()
+            return res
         else:
             try:
-                res = self._build(force)
-            finally:
+                res = self._run()
+            except Exception as e:
+                msg = 'Error building task "{}"' .format(self.name)
+                self._logger.exception(msg)
                 self.exec_status = TaskStatus.Errored
 
-            self.exec_status = TaskStatus.Executed
-            self.product._clear_cached_status()
+                # if there isn't anything left to run, raise exception here
+                if self.on_failure is None:
+                    if isinstance(e, DAGBuildEarlyStop):
+                        raise DAGBuildEarlyStop('Stopping task {} gracefully'
+                                                .format(self.name)) from e
+                    else:
+                        raise TaskBuildError(msg) from e
 
-            return res
+                build_success = False
+                build_exception = e
+            else:
+                build_success = True
+                build_exception = None
 
-    def _build(self, force):
+            if build_success:
+                try:
+                    # this not only runs the hook, but also
+                    # calls save metadata and checks that the product exists
+                    self._run_on_finish()
+                except Exception as e:
+                    self.exec_status = TaskStatus.Errored
+                    msg = ('Exception when running on_finish '
+                           'for task "{}": {}'
+                           .format(self.name, e))
+                    self._logger.exception(msg)
+
+                    if isinstance(e, DAGBuildEarlyStop):
+                        raise DAGBuildEarlyStop('Stopping task {} gracefully'
+                                                .format(self.name)) from e
+                    else:
+                        raise TaskBuildError(msg) from e
+                else:
+                    self.exec_status = TaskStatus.Executed
+
+                return res
+            else:
+                try:
+                    self._run_on_failure()
+                except Exception as e:
+                    msg = ('Exception when running on_failure '
+                           'for task "{}": {}'
+                           .format(self.name, e))
+                    self._logger.exception(msg)
+                    raise TaskBuildError(msg) from e
+
+                if isinstance(build_exception, DAGBuildEarlyStop):
+                    raise DAGBuildEarlyStop(
+                        'Stopping task {} gracefully'
+                        .format(self.name)) from build_exception
+                else:
+                    msg = 'Error building task "{}"' .format(self.name)
+                    raise TaskBuildError(msg) from build_exception
+
+    def _run(self):
+        """
+        Run task if certain status conditions are ok, otherwise raise a
+        TaskBuildError exception
+        """
         # cannot keep running, we depend on the render step to get all the
         # parameters resolved (params, upstream, product)
         if self.exec_status == TaskStatus.WaitingRender:
@@ -422,19 +522,15 @@ class Task(abc.ABC):
             raise TaskBuildError('Attempted to run task "{}", whose '
                                  'status is TaskStatus.Aborted'
                                  .format(self.name))
-        elif self.exec_status == TaskStatus.Skipped and not force:
+        elif self.exec_status == TaskStatus.Skipped:
             raise TaskBuildError('Attempted to run task "{}", whose '
-                                 'status TaskStatus.Skipped. Use force=True '
-                                 'if you want to execute it anyway'
+                                 'status TaskStatus.Skipped. Render again and'
+                                 'set force=True if you want to force '
+                                 'execution'
                                  .format(self.name))
 
         # NOTE: should i fetch metadata here? I need to make sure I have
         # the latest before building
-
-        if force:
-            self._logger.info('Forcing run "%s", status ignored...',
-                              self.name)
-
         self._logger.info('Starting execution: %s', repr(self))
 
         then = datetime.now()
@@ -456,7 +552,9 @@ class Task(abc.ABC):
         """
         Renders code and product, all upstream tasks must have been rendered
         first, for that reason, this method will usually not be called
-        directly but via DAG.render(), which renders in the right order
+        directly but via DAG.render(), which renders in the right order.
+
+        Render fully determines whether a task should run or not.
 
         Parameters
         ----------
@@ -484,15 +582,7 @@ class Task(abc.ABC):
         self.params._dict['product'] = self.product
 
         try:
-            if self.source.needs_render:
-                # if this task has upstream dependencies, render using the
-                # context manager, which will raise a warning if any of the
-                # dependencies is not used, otherwise just render
-                if self.params.get('upstream'):
-                    with self.params.get('upstream'):
-                        self.source.render(self.params)
-                else:
-                    self.source.render(self.params)
+            self.source.render(self.params)
         except Exception as e:
             self.exec_status = TaskStatus.ErroredRender
             raise type(e)('Error rendering source from Task "{}", '
@@ -509,6 +599,8 @@ class Task(abc.ABC):
                 self._exec_status = TaskStatus.Skipped
             else:
                 self._exec_status = TaskStatus.WaitingExecution
+                self._logger.debug('Forcing status "%s", outdated conditions'
+                                   ' ignored...', self.name)
         else:
             all_upstream_done = all([t.exec_status
                                      in {TaskStatus.Executed,
@@ -521,23 +613,15 @@ class Task(abc.ABC):
                 self._exec_status = TaskStatus.Skipped
             else:
                 self._exec_status = TaskStatus.WaitingUpstream
+                self._logger.debug('Forcing status "%s", outdated conditions'
+                                   ' ignored...', self.name)
 
-        self._validate_render()
         self._run_on_render()
 
-    def _validate_render(self):
-        """
-        This hook is executed after rendering, it can be used to perform
-        validation on parameters that might cause runtime errors. e.g.
-        if is used by PythonCallable to check if the function signature
-        matches parameters passed. Note: if is executed before the
-        user-supplied on_render hook
-        """
-        pass
-
     def set_upstream(self, other):
         self.dag._add_edge(other, self)
 
+    # FIXME: delete
     def plan(self):
         """Shows a text summary of what this task will execute
         """
@@ -609,7 +693,7 @@ class Task(abc.ABC):
         data['Product client'] = (repr(self.product.client)
                                   if hasattr(self.product, 'client')
                                   else None)
-        data['Doc (short)'] = self.source.doc_short
+        data['Doc (short)'] = _doc_short(self.source.doc)
         data['Location'] = self.source.loc
 
         return Row(data)
@@ -704,14 +788,6 @@ class Task(abc.ABC):
     def __str__(self):
         return str(self.product)
 
-    def _short_repr(self):
-        def short(s):
-            max_l = 30
-            return s if len(s) <= max_l else s[:max_l - 3] + '...'
-
-        return ('{} -> \n{}'
-                .format(short(self.name), self.product._short_repr()))
-
     def __getstate__(self):
         state = self.__dict__.copy()
         # _logger is not pickable, so we remove them and build
@@ -723,3 +799,10 @@ class Task(abc.ABC):
         self.__dict__.update(state)
         self._logger = logging.getLogger('{}.{}'.format(__name__,
                                                         type(self).__name__))
+
+
+def _doc_short(doc):
+    if doc is not None:
+        return doc.split('\n')[0]
+    else:
+        return None
